/****************************************************************************************************************************
    ___   ___  ___  _   _  ___   ___   ____ ___  ____  
 * / _ \ /___)/ _ \| | | |/ _ \ / _ \ / ___) _ \|    \ 
 *| |_| |___ | |_| | |_| | |_| | |_| ( (__| |_| | | | |
 * \___/(___/ \___/ \__  |\___/ \___(_)____)___/|_|_|_|
 *                  (____/ 
 * OSOYOO Raspberry Pi Pico Robot Car lesson 3 : Obstacle Avoidance
 * Tutorial link:  https://osoyoo.com/?p=49552
 * CopyRIGHT www.osoyoo.com
*/
/*****************************************************************************************************************************/

#define _PWM_LOGLEVEL_        3
#include "RP2040_PWM.h"
#include "Servo.h"
#define IN1      16    // GP16 to IN1 right motor direction pin
#define IN2      17    // GP17 to IN2 right motor direction pin
#define IN3      18    // GP18 to IN3 left motor direction pin
#define IN4      19    // GP19 to IN4 left motor direction pin
#define ENA      9     // GP10 to ENA right motor PWM
#define ENB      10    // GP9  to ENB left motor PWM
#define Echo_PIN  7    // Ultrasonic Echo pin to G7
#define Trig_PIN  8    // Ultrasonic Trig pin to G8
#define SERVO_PIN 15   // servo connect to G15
#define LPT 2 // scan loop coumter
RP2040_PWM* PWM_Left;
RP2040_PWM* PWM_Right;

int leftscanval, centerscanval, rightscanval, ldiagonalscanval, rdiagonalscanval;
const int distancelimit = 20; //distance limit for obstacles in front           
const int sidedistancelimit = 20; //minimum distance in cm to obstacles at both sides (the car will allow a shorter distance sideways)
int distance;
int numcycles = 0;
const int turntime = 400; //Time the robot spends turning (miliseconds)
const int backtime = 400; //Time the robot spends turning (miliseconds)
int thereis;
Servo head;

float frequency;
float SPEED=50.f;
float TURN_SPEED=100.f;
void set_speed(float speed_left,float speed_right )
{
  PWM_Left->setPWM(ENB, frequency, speed_left, true);
  PWM_Right->setPWM(ENA, frequency, speed_right, true);
}
void stopcar( )
{
  PWM_Left->setPWM(ENB, frequency, 0, true);
  PWM_Right->setPWM(ENA, frequency, 0, true);
  digitalWrite(IN1,0);
  digitalWrite(IN2,0);
  digitalWrite(IN3,0);
  digitalWrite(IN4,0);
  
}
void  backward()
{
  digitalWrite(IN1,HIGH);
  digitalWrite(IN2,LOW);
  digitalWrite(IN3,HIGH);
  digitalWrite(IN4,LOW);
  set_speed(SPEED,SPEED);
}
void forward()
{
  digitalWrite(IN1, LOW);
  digitalWrite(IN2,HIGH);
  digitalWrite(IN3, LOW);
  digitalWrite(IN4,HIGH);
  set_speed(SPEED,SPEED);
}
void left_turn()
{
  digitalWrite(IN1,LOW);
  digitalWrite(IN2,HIGH);
  digitalWrite(IN3,HIGH);
  digitalWrite(IN4,LOW);
  set_speed(0,TURN_SPEED);
}
void right_turn()
{
  digitalWrite(IN1,HIGH);
  digitalWrite(IN2,LOW);
  digitalWrite(IN3,LOW);
  digitalWrite(IN4,HIGH);
  set_speed(TURN_SPEED,0);
}
int watch(){
  long echo_distance;
  digitalWrite(Trig_PIN,LOW);
  delayMicroseconds(5);                                                                              
  digitalWrite(Trig_PIN,HIGH);
  delayMicroseconds(15);
  digitalWrite(Trig_PIN,LOW);
  echo_distance=pulseIn(Echo_PIN,HIGH);
  echo_distance=echo_distance*0.01657; //how far away is the object in cm
  //Serial.println((int)echo_distance);
  return round(echo_distance);
}
String watchsurrounding(){
/*  obstacle_status is a binary integer, its last 5 digits stands for if there is any obstacles in 5 directions,
 *   for example B101000 last 5 digits is 01000, which stands for Left front has obstacle, B100111 means front, right front and right ha
 */
 
int obstacle_status =B100000;
  centerscanval = watch();
  if(centerscanval<distancelimit){
   // stopcar();
    
    obstacle_status  =obstacle_status | B100;
    }
  head.write(120);
  delay(100);
  ldiagonalscanval = watch();
  if(ldiagonalscanval<distancelimit){
  //  stopcar();
    
     obstacle_status  =obstacle_status | B1000;
    }
  head.write(170); //Didn't use 180 degrees because my servo is not able to take this angle
  delay(300);
  leftscanval = watch();
  if(leftscanval<sidedistancelimit){
  //  stopcar();
    
     obstacle_status  =obstacle_status | B10000;
    }

  head.write(90); //use 90 degrees if you are moving your servo through the whole 180 degrees
  delay(100);
  centerscanval = watch();
  if(centerscanval<distancelimit){
   // stopcar();
    
    obstacle_status  =obstacle_status | B100;
    }
  head.write(40);
  delay(100);
  rdiagonalscanval = watch();
  if(rdiagonalscanval<distancelimit){
   // stopcar();
    
    obstacle_status  =obstacle_status | B10;
    }
  head.write(0);
  delay(100);
  rightscanval = watch();
  if(rightscanval<sidedistancelimit){
   // stopcar();
    
    obstacle_status  =obstacle_status | 1;
    }
  head.write(90); //Finish looking around (look forward again)
  delay(300);
   String obstacle_str= String(obstacle_status,BIN);
  obstacle_str= obstacle_str.substring(1,6);
  
  return obstacle_str; //return 5-character string standing for 5 direction obstacle status
}
void auto_avoidance(){

  ++numcycles;
  if(numcycles>=LPT){ //Watch if something is around every LPT loops while moving forward 
     stopcar();
    String obstacle_sign=watchsurrounding(); // 5 digits of obstacle_sign binary value means the 5 direction obstacle status
      Serial.print("begin str=");
      Serial.println(obstacle_sign);
      if( obstacle_sign=="10000"){
      Serial.println("SLIT right");
      set_speed(TURN_SPEED,SPEED);
      forward();
 
      delay(turntime);
      stopcar();
    }
     else    if( obstacle_sign=="00001"  ){
     Serial.println("SLIT LEFT");
     set_speed(SPEED,TURN_SPEED);
     forward();
  
     delay(turntime);
     stopcar();
    }
    else if( obstacle_sign=="11100" || obstacle_sign=="01000" || obstacle_sign=="11000"  || obstacle_sign=="10100"  || obstacle_sign=="01100" ||obstacle_sign=="00100"  ||obstacle_sign=="01000" ){
     Serial.println("hand right");
     right_turn();
     set_speed(TURN_SPEED,TURN_SPEED);
     delay(turntime);
     stopcar();
    } 
    else if( obstacle_sign=="00010" || obstacle_sign=="00111" || obstacle_sign=="00011"  || obstacle_sign=="00101" || obstacle_sign=="00110" || obstacle_sign=="01010" ){
    Serial.println("hand left");
    left_turn();//Turn left
    set_speed(TURN_SPEED,TURN_SPEED);
    delay(turntime);
    stopcar();
    }
 
    else if(  obstacle_sign=="01111" ||  obstacle_sign=="10111" || obstacle_sign=="11111"  ){
    Serial.println("hand back left");
    backward();
    set_speed( SPEED,SPEED);
    delay(backtime);
    stopcar();
        } 
    else if( obstacle_sign=="11011"  ||    obstacle_sign=="11101"  ||  obstacle_sign=="11110"  || obstacle_sign=="01110"  ){
    Serial.println("hand back right");
    backward();
    set_speed(SPEED,SPEED);
    delay(backtime);
    stopcar();
        }    
  
    else Serial.println("no handle");
    numcycles=0; //Restart count of cycles
  } else {
    set_speed(SPEED,SPEED);
    forward();  // if nothing is wrong go forward using go() function above.
    delay(backtime);
    stopcar();
  }
  
  //else  Serial.println(numcycles);
  
  distance = watch(); // use the watch() function to see if anything is ahead (when the robot is just moving forward and not looking around it will test the distance in front)
  if (distance<distancelimit){ // The robot will just stop if it is completely sure there's an obstacle ahead (must test 25 times) (needed to ignore ultrasonic sensor's false signals)
 Serial.println("final go back");
  backward();
  set_speed(SPEED,TURN_SPEED);
  delay(backtime);
      ++thereis;}
  if (distance>distancelimit){
      thereis=0;} //Count is restarted
  if (thereis > 25){
  Serial.println("final stop");
    stopcar(); // Since something is ahead, stop moving.
    thereis=0;
  }
}
void setup()
{
  pinMode(Trig_PIN,OUTPUT); 
  pinMode(Echo_PIN,INPUT); 
  pinMode(IN1,OUTPUT);
  pinMode(IN2,OUTPUT);
  pinMode(IN3,OUTPUT);
  pinMode(IN4,OUTPUT);
  frequency = 1000;
  PWM_Left = new RP2040_PWM(ENA, frequency, 50);
  PWM_Right = new RP2040_PWM(ENB,frequency, 50);
  if (PWM_Left)
  {
    PWM_Left->setPWM();
  }
  if (PWM_Right)
  {
    PWM_Right->setPWM();
  }
  head.attach(SERVO_PIN); 
  head.write(135);
  delay(1000);
    head.write(45);
  delay(1000);
    head.write(90);
  delay(2000);
}

 
void loop()
{
 
  auto_avoidance();
}
